@@ -2,11 +2,13 @@ try:
     from simplejson import dumps, loads
 except ImportError:  # pragma: no cover
     from json import dumps, loads  # noqa
-from itertools import chain
+from inspect import Arguments
+from itertools import chain, tee
 from mimetypes import guess_type, add_type
 from os.path import splitext
 import logging
 import operator
+import types
 
 import six
 
@@ -30,12 +32,14 @@ logger = logging.getLogger(__name__)
 
 class RoutingState(object):
 
-    def __init__(self, request, response, app, hooks=[], controller=None):
+    def __init__(self, request, response, app, hooks=[], controller=None,
+                 arguments=None):
         self.request = request
         self.response = response
         self.app = app
         self.hooks = hooks
         self.controller = controller
+        self.arguments = arguments
 
 
 class Request(WebObRequest):
@@ -325,6 +329,7 @@ class PecanBase(object):
         passed the argument specification for the controller.
         '''
         args = []
+        varargs = []
         kwargs = dict()
         valid_args = argspec.args[1:]  # pop off `self`
         pecan_state = state.request.pecan
@@ -353,7 +358,7 @@ class PecanBase(object):
         if [i for i in remainder if i]:
             if not argspec[1]:
                 abort(404)
-            args.extend(remainder)
+            varargs.extend(remainder)
 
         # get the default positional arguments
         if argspec[3]:
@@ -376,7 +381,7 @@ class PecanBase(object):
                 if name not in argspec[0]:
                     kwargs[encode_if_needed(name)] = value
 
-        return args, kwargs
+        return args, varargs, kwargs
 
     def render(self, template, namespace):
         renderer = self.renderers.get(
@@ -491,9 +496,6 @@ class PecanBase(object):
             )
             raise exc.HTTPNotFound
 
-        # handle "before" hooks
-        self.handle_hooks(self.determine_hooks(controller), 'before', state)
-
         # fetch any parameters
         if req.method == 'GET':
             params = dict(req.GET)
@@ -501,15 +503,19 @@ class PecanBase(object):
             params = dict(req.params)
 
         # fetch the arguments for the controller
-        args, kwargs = self.get_args(
+        args, varargs, kwargs = self.get_args(
             state,
             params,
             remainder,
             cfg['argspec'],
             im_self
         )
+        state.arguments = Arguments(args, varargs, kwargs)
 
-        return controller, args, kwargs
+        # handle "before" hooks
+        self.handle_hooks(self.determine_hooks(controller), 'before', state)
+
+        return controller, args+varargs, kwargs
 
     def invoke_controller(self, controller, args, kwargs, state):
         '''
@@ -565,15 +571,38 @@ class PecanBase(object):
             resp.text = result
         elif result:
             resp.body = result
-        elif response.status_int == 200:
-            resp.status = 204
 
-        if resp.status_int in (204, 304):
-            resp.content_type = None
-        elif pecan_state['content_type']:
+        if pecan_state['content_type']:
             # set the content type
             resp.content_type = pecan_state['content_type']
 
+    def _handle_empty_response_body(self, state):
+        # Enforce HTTP 204 for responses which contain no body
+        if state.response.status_int == 200:
+            # If the response is a generator...
+            if isinstance(state.response.app_iter, types.GeneratorType):
+                # Split the generator into two so we can peek at one of them
+                # and determine if there is any response body content
+                a, b = tee(state.response.app_iter)
+                try:
+                    next(a)
+                except StopIteration:
+                    # If we hit StopIteration, the body is empty
+                    state.response.status = 204
+                finally:
+                    state.response.app_iter = b
+            else:
+                text = None
+                if state.response.charset:
+                    # `response.text` cannot be accessed without a charset
+                    # (because we don't know which encoding to use)
+                    text = state.response.text
+                if not any((state.response.body, text)):
+                    state.response.status = 204
+
+        if state.response.status_int in (204, 304):
+            state.response.content_type = None
+
     def __call__(self, environ, start_response):
         '''
         Implements the WSGI specification for Pecan applications, utilizing
@@ -643,6 +672,8 @@ class PecanBase(object):
                 self.determine_hooks(state.controller), 'after', state
             )
 
+        self._handle_empty_response_body(state)
+
         # get the response
         return state.response(environ, start_response)
 
@@ -671,11 +702,11 @@ class ExplicitPecan(PecanBase):
         except IndexError:
             raise signature_error
 
-        args, kwargs = super(ExplicitPecan, self).get_args(
+        args, varargs, kwargs = super(ExplicitPecan, self).get_args(
             state, all_params, remainder, argspec, im_self
         )
         args = [state.request, state.response] + args
-        return args, kwargs
+        return args, varargs, kwargs
 
 
 class Pecan(PecanBase):
@@ -727,12 +758,14 @@ class Pecan(PecanBase):
             state.hooks = []
             state.app = self
             state.controller = None
+            state.arguments = None
             return super(Pecan, self).__call__(environ, start_response)
         finally:
             del state.hooks
             del state.request
             del state.response
             del state.controller
+            del state.arguments
             del state.app
 
     def init_context_local(self, local_factory):
@@ -746,6 +779,7 @@ class Pecan(PecanBase):
         state.response = _state.response
         controller, args, kw = super(Pecan, self).find_controller(_state)
         state.controller = controller
+        state.arguments = _state.arguments
         return controller, args, kw
 
     def handle_hooks(self, hooks, *args, **kw):
